/*
  Copyright Teppo Kurki, ????
  
  Sailgauge svg based component for showing sail information.
  Component is configurable by .css with class name.
  Methods:
    sailgauge(options) - constructor
        options        - object of optional parameters. Define it as {ID:'sailgauge',className:'sailgauge2'}
                         Optional parameters are;
          ID                  - svg group id. So this can be searched with d3.select(#<ID>);
          className           - defines class name for .css . Default 'compass'
          margin              - margin around rose box. Default 80
          mainTickSize        - Size of 10 degree tics. Default 20.
          minorTickSize       - Size of 5 degree tics. Default 10.
          trackFilter         - Minimum value for rotating rose. This prevents small jitter. Default 0.5.
          laylineAngle        - angle of laylines. Default 30.
          speedLabelPos       - Object for position for speed label. Default {x:'center',y:'center',dx:0,dy:0}.
          trackLabelPos       - Object for position for speed label. Default {x:'center',y:'above:#roseOuter',dx:0,dy:0}
          courseShowStyle     - Select how course/heading will be shown. Default is 1
                                  - 1 course (COG) up. Boat will be tilted according to drift.
                                  - 2 heading up. COG will be shown with courseIndicator (see styles)
          useAnimatedRotation - Use animation in rotation. Default is false. See my comment "About animated rotation"
          
    init(svgContainer,x,y,boxWidth,scale) - initialize and draw componenent
        svgContainer   - svg container
        options        - object of optional parameters. Define it as {center:{x:200, y:200}}
                         Optional parameters are;
          center:{x,y} - Object of center of rose. Define as center:{x:500,y:500}. 
                         These are container coordinates. Default is center of container.
          size         - size of sail gauge area in container units. Default is min size of container.
          
    updateSpeedOverGround(speed, updateTrueWind) - Function updates speed over ground and true wind if updateTrueWind==true
        speed           - numeric speed over ground in knots
        updateTrueWind  - true or false. Default false;
                      
    updateApparentWind(speed, angle, updateTrueWind) - Function updates apparent wind and true wind if updateTrueWind==true
        speed           - numeric speed of apparent wind in m/s
        angle           - numeric angle of apparent wind in degrees
        updateTrueWind  - true or false. Default false;
        
    updateApparentWindAngle(angle, updateTrueWind) - Function updates apparent wind angle and true wind if updateTrueWind==true
        angle           - numeric angle of apparent wind in degrees
        updateTrueWind  - true or false. Default false;
        
    updateApparentWindSpeed(speed, updateTrueWind) - Function updates apparent wind speed and true wind if updateTrueWind==true
        speed           - numeric speed of apparent wind in m/s
        updateTrueWind  - true or false. Default false;
        
    updateTrueWindAngle(angle) - Function updates true wind angle.
        angle           - numeric angle of true wind in degrees
        
    updateTrueWindSpeed(speed) - Function updates true wind speed
        speed           - numeric speed of true wind in m/s

    updateMark(msg) - Function updates bearing and distance to waypoint
        msg             - object of {distance:<distance>,bearing:<bearing>}. 
                          Distance in nautical miles
                          Bearing in degrees
    
    updateCourse(value)- Function updates boat course over ground.
        value           - numeric course over ground in degrees
        
    updateDepthDisplay(depth) - Function updates depth. 
        value           - numeric depth in meters. Is is your decicion to use depth below keel or water depth.
    
  Properties:
    All contructor options and -
    svgContainer       - svgContainer where node stays
    scale              - scale of the componenent
    center             - Object of center of the rose. center has properties x,y.
    SpeedOverGround    - current value for speed over ground in knots
    ApparentWindAngle  - current value for apparent wind angle in degrees
    ApparentWindSpeed  - current value for apparent wind speed in m/s 
    TrueWindAngle      - current value for true wind angle in degrees 
    TrueWindSpeed      - current value for true wind speed in m/s
    Depth              - current value for depth in meters
    trackTrue          - current value for course over ground 
  
  Styles, using default class name "compass" (see className property)
    .sailgauge {
    }
    .sailgauge .rose {stroke:#A63400;fill:#679fd2}
    .sailgauge .rose .text{stroke:#A63400;fill:#A63400}
    .sailgauge .apparentwind{fill:#043a6b;stroke:white}
    .sailgauge .apparentwind .text{stroke:white;fill:white ;font-size:260%}
    .sailgauge .truewind{fill:#679fd2;stroke:black}
    .sailgauge .truewind .text{stroke:black;fill:black ;font-size:260%}
    .sailgauge .positivegaugetext {stroke:#679fd2;fill:#679fd2}
    .sailgauge .speedtextbox {stroke:#679fd2;fill:#679fd2}
    .sailgauge .speedtext {stroke:#black;fill:black ;font-size:500%}
    .sailgauge .tracktextbox {stroke:#679fd2;fill:#679fd2}
    .sailgauge .tracktext {stroke:#black;fill:black ;font-size:500%}
    .sailgauge .sparkline {stroke:#679fd2}
    .sailgauge .boat {fill:#26537c}
    .sailgauge .mark {fill:red;stroke:none}
    .sailgauge .marktext {fill:black;stroke:black;font-size:200%}
    .sailgauge .markBackground {fill:white;stroke:none}
    .sailgauge .rose .OuterCircle{stroke:red; stroke-width:2; fill:gray}
    .sailgauge .rose .MainTick{stroke:A63400; stroke-width:2; fill:A63400}
    .sailgauge .rose .MinorTick{stroke:A63400; stroke-width:1; fill:A63400}
    .sailgauge .rose .CardinalText{fill:#A63400; stroke-width:1 ;font-size:300%}
    .sailgauge .rose .IntercardinalText{stroke:#A63400; stroke-width:1 ;fill:#A63400;font-size:240%}
    .sailgauge .courseIndicator{stroke:blue; stroke-width:6; fill:blue}
    .sailgauge.text {font-size:60px}

  Sample usage:
  ...
	<body>
    ...
    <div id="DisplayContainer" class="svg"></div>
    ...
    <script>
      var svgContainer=AddSvgContainer('#DisplayContainer',1000);
      var sailgauge = new SailGauge({trackLabelPos:{x:'left',y:'below:#speedLabel',dx:0, dy:5},laylineAngle:35});
      sailgauge.init(svgContainer,{center:{x:500,y:500},size:800});
  
  Aligning elements:
    You can set some element Aligning on construction. You simply define with constructor options the element x,y
    position e.g. {trackLabelPos:{x:'center',y:'below:#speedLabel',dx:0,dy:10}}
    
    The sailgauge coordinate system is in it's own transform, which box is -400,-400 -> 400,400. So the center of
    component is 0,0.
    
    Next element positions can be defined:
      speedLabelPos
      trackLabelPos
      
    Possible positioning referencies are:
      left:<val>      - align left of given value. val can be fixed value like 100 or
                        element reference like #speedLabel . 
                        If :<val> is missing, default is sailgauge.
      top:<val>       - align top of given value.
      right:<val>     - align right of given value.
      bottom:<val>    - align bottom of given value.
      center:<val>    - align center of given value.
      above:<val>     - align above given value.     
      below:<val>     - align below given value.    

    Possible element referencies are:
      speedLabel
      trackLabel
      roseOuter
      roseInner
      sailgaugeRect
      
  About animated rotation:
    Animated rotation is nice feature, but unfortunately eats performance. See values below
    for 5 upgrades of Heading, SOG, wind
      Animated rotation on
      Total: 2.50 s
        39.1 msScripting
        56.3 msRendering
        48.3 msPainting
        57.4 msOther
       201.1 msRun    8,03%
      2302.5 msIdle
      2503.6 msTotal

      Animated rotation off
      Total: 2.51 s
        10.6 msScripting
         5.2 msRendering
         3.6 msPainting
         9.9 msOther
        29.3 msRun     1.16%
      2477.5 msIdle
      2506.8 msTotal
    
*/

function SailGauge(options) {
  var def={ID:null
           ,className:'sailgauge'
           ,margin:80
           ,mainTickSize:20
           ,minorTickSize:10
           ,trackFilter:0.5
           ,laylineAngle:30
           ,speedLabelPos:{x:'center',y:'center',dx:0,dy:0}
           ,trackLabelPos:{x:'center',y:'above:#roseOuter',dx:0,dy:0}
           ,courseShowStyle:1
           ,useAnimatedRotation:false
          }; 
  if (options==null) options=def;
  CombineDefaultsToObject(options,def);
  CombineDefaultsToObject(this,options);

  if ( this.ID==null ) this.ID='Sailgauge';
  
  this.visible = true;
  this.SpeedOverGround = 0;
  this.ApparentWindAngle = 0;
  this.ApparentWindSpeed = 0;
  this.TrueWindAngle = 0;
  this.TrueWindSpeed = 0;
  this.Heading = 0;
  this.trackTrue=0;
  this.trackTrueValid=false; // do we have valid value
  this.Drift=0;
  this.Depth = 0;
  this.bearingToMark=0;
  
  this.dispatchers =  {};
  
  this.margin=DefaultNum(options.margin,def.margin);
  this.mainTickSize=DefaultNum(options.mainTickSize,def.mainTickSize);
  this.minorTickSize=DefaultNum(options.minorTickSize,def.minorTickSize);
  this.trackFilter=DefaultNum(options.trackFilter,def.trackFilter);
  this.laylineAngle=DefaultNum(options.laylineAngle,def.laylineAngle);

  this.r=400;
  this.innerR=this.r-this.margin-this.mainTickSize;
}

//************************************************************************************
SailGauge.prototype = {
  init: function (svgContainer,options) {
    var def={ID:null
            ,center:{x:null,y:null} 
            ,size:null
            }; 
    if (options==null) options=def;
    options=CombineDefaultsToObject(options,def);

    if ( typeof svgContainer == "string" ) { // Container defined by string, create default
      svgContainer=AddSvgContainer(svgContainer,{width:800}); 
    }

    this.svgContainer=svgContainer;

    var svgSizeInfo=GetSvgSizeInfo(svgContainer);
    options.center.x=DefaultNum(options.center.x,svgSizeInfo.width/2);
    options.center.y=DefaultNum(options.center.y,svgSizeInfo.height/2);
    var defSize=Math.min(svgSizeInfo.width,svgSizeInfo.height);
    options.size=DefaultNum(options.size,defSize);
    this.scale=options.size/2/this.r;
    this.center=options.center;
    
    this.drawSvg(svgContainer, 800);
    this.startPeriodicTask();
  },
  //************************************************************************************
  calcTrueWind: function() {
    var trueWind=CalcTrueWind(this.ApparentWindSpeed,DegToRad(this.ApparentWindAngle)
                             ,KnToms(this.SpeedOverGround),DegToRad(this.Drift));
    this.updateCalculatedTrueWindSpeed(trueWind.speed);
    this.updateCalculatedTrueWindAngle(RadToDeg(trueWind.angle));
  },
  //************************************************************************************
  updateSpeedOverGround: function(speed) {
    if ( speed==null ) speed="N/A";
    if ( !isNaN(speed) ) {
      this.SpeedOverGround=speed;
      this.svgSpeedLabel.text(speed.toFixed(1));
      
    } else {
      this.SpeedOverGround=0;
      this.svgSpeedLabel.text(speed);
    }
    this.lastSpeedReceiveTime = Date.now();
  },
  //************************************************************************************
  updateApparentWind: function (speed, angle) {
    this.updateApparentWindSpeed(speed,false);
    this.updateApparentWindAngle(angle,false);
    
    if (Date.now() - this.lastTrueWindReceiveTime > 3 * 1000) { // if we have not received true wind for a while, use calculated value
      if ( speed!=null && !isNaN(speed) && angle!=null && !isNaN(angle) ) {
        this.calcTrueWind();
      } else {
        this.updateTrueWindSpeed("Lost");
      }
    }
  },
  //************************************************************************************
  updateApparentWindAngle: function (angle, updateTrueWind) {
    if ( angle!=null && !isNaN(angle) ) {
      var rotAdd=this.Drift;
      switch ( this.courseShowStyle ) {
        case 2:
          rotAdd=0;
          break;
      }
      this.rotatesvgAnimated(this.svgApparentWindMarker, this.ApparentWindAngle+rotAdd, angle+rotAdd, 0, 0, 200);
      this.ApparentWindAngle=angle;
      this.svgApparentWindmarkerText.attr("transform", "rotate(" + (-angle-rotAdd) + " 0 " + this.apparentWindMarkerTextCenter + ")");

      if (updateTrueWind == null || updateTrueWind==true) {
        if (Date.now() - this.lastTrueWindReceiveTime > 3 * 1000) { // if we have not received true wind for a while, use calculated value
          if ( angle!=null && !isNaN(angle) ) {
            this.calcTrueWind();
          } else {
            this.updateTrueWindSpeed("Lost");
          }
        }
      }
    }
  },
  //************************************************************************************
  updateApparentWindSpeed: function (speed, updateTrueWind) {
    if ( speed==null ) speed="N/A";
    if ( !isNaN(speed) ) {
      this.ApparentWindSpeed = speed;
      this.svgApparentWindmarkerText.text(Number(speed).toFixed(1));

    } else {
      this.ApparentWindSpeed = 0;
      this.svgApparentWindmarkerText.text(speed);
    }

    if (updateTrueWind == null || updateTrueWind==true) {
      if (Date.now() - this.lastTrueWindReceiveTime > 3 * 1000) { // if we have not received true wind for a while, use calculated value
        if ( speed!=null && !isNaN(speed) ) {
          this.calcTrueWind();
        } else {
          this.updateTrueWindSpeed("Lost");
        }
      }
    }

    this.svgApparentWindMarker.attr("display","inline")
    this.lastApparentWindReceiveTime = Date.now();
  },
  //************************************************************************************
  updateCalculatedTrueWindAngle: function (angle) {
    if ( angle!=null && !isNaN(angle) ) {
      this.rotateAnimated('#windmarker', this.TrueWindAngle, angle, 0, 0, 200);
      this.rotateAnimated('#laylines', this.TrueWindAngle, angle, 0, 0, 200);
      this.TrueWindAngle=angle;
      this.svgWindmarkerText.attr("transform", "rotate(" + (-angle) + " 0 " + this.trueWindMarkerTextCenter + ")")
    }
  },
  //************************************************************************************
  updateTrueWindAngle: function (angle) {
    if ( angle!=null && !isNaN(angle) ) {
      this.updateCalculatedTrueWindAngle(angle);
      this.lastTrueWindReceiveTime = Date.now();
    }
  },
  //************************************************************************************
  updateCalculatedTrueWindSpeed: function (speed) {
    this.TrueWindSpeed=speed;
    this.svgWindmarkerText.text(Number(speed).toFixed(1));
    this.svgTrueWindMarker.attr("display","inline");
    this.laylines.attr("display","inline");
  },
  //************************************************************************************
  updateTrueWindSpeed: function (speed) {
    if ( speed==null ) speed="N/A";
    if ( speed!=null && !isNaN(speed) ) {
      this.updateCalculatedTrueWindSpeed();
      this.lastTrueWindReceiveTime = Date.now();
    } else {
      if (Date.now() - this.lastTrueWindReceiveTime > 3 * 1000) {
        this.svgWindmarkerText.text(speed);
      }
    }
  },
  //************************************************************************************
  updateMarkTextRotation: function () {
    if ( this.trackTrueValid ) {
      this.svgMarkText.attr("transform", "rotate(" + (this.bearingToMark + this.trackTrue) + " 0 "+this.markTextCenter+")");
      this.svgMarkDistance.attr("transform", "rotate(" + (this.bearingToMark + this.trackTrue) + " 0 "+this.markDistanceTextCenter+")");
    }
  },
  //************************************************************************************
  updateMark: function (msg) {
    this.svgMarkDistance.text(Number(msg.distance).toFixed(1));
    this.svgMark.attr("display", "inline");
    this.svgMarkText.text(msg.bearing.toFixed(0));
    this.rotatesvgAnimated(this.svgMark, -this.bearingToMark, -msg.bearing, 0, 0, 200);
    this.bearingToMark = msg.bearing;
    this.lastAutopilotReceiveTime = Date.now();
    this.updateMarkTextRotation();
  },
  //************************************************************************************
  updateHeading: function (value) {
    if ( value!=null && !isNaN(value) && this.trackTrueValid ) this.Drift=value-this.trackTrue;
  },
  //************************************************************************************
  updateCourse: function (value) {
    if ( value==null ) value="N/A";
    if ( !isNaN(value) ) {
      this.svgTrackLabel.text(value.toFixed(0) + '°');
      if ( this.trackTrueValid ) {
        this.rotatesvgAnimated(this.svgRose, -this.trackTrue, -value, 0, 0, 200);
       
        switch ( this.courseShowStyle ) {
          case 2:
            this.svgCourceIndicator.attr("transform", "rotate("+(this.Drift)+" 0 0 )");
            break;
          default:
            this.svgBoat.attr("transform", "rotate("+(this.Drift)+" 0 0 )");
        }
        this.svgBoat.attr("display","inline");
        this.updateMarkTextRotation();
      }
      this.trackTrue = value;
      this.trackTrueValid = true;
    } else {
      this.trackTrueValid=false;
      this.svgTrackLabel.text(value);
    }
    this.lastCourceReceiveTime = Date.now();
  },
  lastDepthReceiveTime: 0,
  lastAutopilotReceiveTime: 0,
  lastSpeedReceiveTime: 0,
  lastCourceReceiveTime: 0,
  lastApparentWindReceiveTime: 0,
  lastTrueWindReceiveTime: 0,
  //************************************************************************************
  updateDepthDisplay: function (depth) {
    if ( depth==null ) depth="N/A";
    if ( !isNaN(depth) ) {
      this.svgDepthText.text(depth.toFixed(1) + 'm').attr("display", "inline");
      var fontSize = this.depthFontSize(depth);
      this.svgDepthText.attr('font-size', fontSize);
      this.svgDepthText.attr('dy', fontSize - 60);
      this.svgDepthText.attr('fill', depth < 6 ? 'red' : null);
      this.svgDepthText.attr('font-weight', depth < 6 ? 'bold' : null);
      this.lastDepthReceiveTime = Date.now();
      this.Depth=depth;
    } else {
      this.Depth=0;
      this.svgDepthText.text(depth);
    }
  },
  //************************************************************************************
  startPeriodicTask: function () {
    var that = this;
    var Now=Date.now();
    setInterval(function hideInactiveElements() {
      if (Now - that.lastAutopilotReceiveTime > 60 * 1000) {
        that.svgMark.attr("display", "none");
        that.lastAutopilotReceiveTime=Now+3600000; // avoid continous object update
      }
      if (Now - that.lastDepthReceiveTime > 5 * 1000) {
        that.svgDepthText.attr("display", "none");
        that.lastDepthReceiveTime=Now+3600000; // avoid continous object update
      }
      if (Now - that.lastApparentWindReceiveTime > 5 * 1000) {
        that.svgApparentWindmarkerText.text("Lost");
        that.lastApparentWindReceiveTime=Now+3600000; // avoid continous object update
        if (Now - that.lastTrueWindReceiveTime > 5 * 1000) {
          that.svgWindmarkerText.text("Lost");
        }
      }
      if (Now - that.lastCourceReceiveTime > 2 * 1000) {
        that.svgTrackLabel.text("Lost");
        that.lastCourceReceiveTime=Now+3600000; // avoid continous object update
      }
      if (Now - that.lastSpeedReceiveTime > 2 * 1000) {
        that.svgSpeedLabel.text("Lost");
        that.lastSpeedReceiveTime=Now+3600000; // avoid continous object update
      }
    }, 2000);
  },
  //************************************************************************************
  setVisible: function(flag) {
    this.visible = flag;
  },
  //************************************************************************************
  SubID: function (ID) {
    return this.ID+ID;
  },
  //************************************************************************************
  Select: function (ID) {
    return d3.select("#"+this.SubID(ID));
  },
  //************************************************************************************
  drawLaylines: function (chart) {
    this.laylines = chart.append("g").attr("display","none")
      .attr("id", "laylines")
      .attr("class", "truewind");
    this.laylines.append("path").attr("d", "M -10,0 l 10,"+(-this.innerR)+" 10,"+(this.innerR)+" z")
      .attr("style", "fill:#00FF00;stroke:#00FF00;stroke-width:1")
      .attr("transform", "rotate("+this.laylineAngle+" 0 0)");
    this.laylines.append("path").attr("d", "M -10,0 l 10,"+(-this.innerR)+" 10,"+(this.innerR)+" z")
      .attr("style", "fill:#ff0000;stroke:#ff0000;stroke-width:1")
      .attr("transform", "rotate("+(-this.laylineAngle)+" 0 0)");
  },
  //************************************************************************************
  drawTrueWindMarker: function (chart) {
    var Text;
    var bbox;
    
    this.svgTrueWindMarker = chart.append("g")
      .attr("id", "windmarker").attr("class", "truewind");
    Text=this.svgTrueWindMarker.append("text")
      .attr("id", "windmarkertext")
      .attr("x", "0").attr("y", -this.innerR)
      .attr("text-anchor", "middle") //.attr("dominant-baseline", "middle")
      .attr('dy', '.3em') // IE does not support 'dominant-baseline' so use dy instead
      .attr("font-size", this.windMarkerFontSize)
      .attr("class", "text")
      .text("i00.0i");
    this.svgWindmarkerText=Text;
    bbox=Text.node().getBBox();
    Text.text("0.0");
    this.trueWindMarkerTextCenter=-(this.innerR-30-bbox.width/2);
    Text.attr("y", this.trueWindMarkerTextCenter);
    this.svgTrueWindMarker.insert("path","#windmarkertext")
      .attr("d", "M 0,"+(-this.innerR)+" l  -15,30  a "+bbox.width/2+","+bbox.width/2+" 0 1,0 30,0 z");
    this.svgTrueWindMarker.attr("display","none");
  },
  //************************************************************************************
  drawApparentWindMarker: function (chart) {
    var Text;
    var bbox;
    
    this.svgApparentWindMarker = chart.append("g")
      .attr("id", "apparentwindmarker")
      .attr("class", "apparentwind");
    Text=this.svgApparentWindMarker.append("text")
      .attr("id", "apparentwindmarkertext")
      .attr("x", "0").attr("y", -this.innerR)
      .attr("text-anchor", "middle") //.attr("dominant-baseline", "middle")
      .attr('dy', '.3em') // IE does not support 'dominant-baseline' so use dy instead
      .attr("class", "text")
      .text("i00.0i");
    this.svgApparentWindmarkerText=Text;
    bbox=Text.node().getBBox();
    Text.text("0.0");
    this.apparentWindMarkerTextCenter=-(this.innerR+30+bbox.width/2);
    Text.attr("y", this.apparentWindMarkerTextCenter);
    this.svgApparentWindMarker.insert("path","#apparentwindmarkertext")
      .attr("d", "M 0,"+(-this.innerR)+" l  -15,-30  a "+bbox.width/2+","+bbox.width/2+" 0 1,1 30,0 z");
    this.svgApparentWindMarker.attr("display","none")
  },
  //************************************************************************************
  drawCompassRose: function (chart, size) {
    var rose = chart.append('g').attr('id', this.SubID('Rose')).attr('class', 'rose');
    rose.append('g').attr('id', 'edge').attr('class', 'p1');
    rose.append('circle').attr('id','roseOuter').attr('cx', 0).attr('cy', 0).attr('r', size / 2 - this.margin).attr('class', 'OuterCircle');
    rose.append('path').attr('id','roseInner')
      .attr('d', "M "+(this.innerR)+" 0 A -100 -100 0 1 1  "+(-this.innerR)+",0 A -100 -100 0 1 1  "+(this.innerR)+" 0 z")
      .attr('class', 'p-1')
      .attr('fill', 'url(#rosegradient)')
      .attr('opacity', '0.8')
      .attr('stroke-width', '2px');
    rose.append('g').attr('id', 'tickmarks');
    rose.append('g').attr('id', 'tickmarksTop');
    return rose;
  },
  //************************************************************************************
  drawMark: function (rose) {
    var Text;
    var bbox;

    var mark = rose.append('g').attr('id', 'mark');
    this.svgMark=mark;
    Text=mark.append('text')
      .attr('id','markdistance')
      .attr('x', '0').attr('y', -this.innerR)
      .attr('class', 'marktext')
      .attr('text-anchor', 'middle')
      .attr('dy', '.35em') // IE does not support 'dominant-baseline' so use dy instead
      .text('00.0i');
    this.svgMarkDistance=Text;
    bbox=Text.node().getBBox();
    var triangleHalfLen=(bbox.width/2)/Math.tan(30*Math.PI/180);
    var triangleHeight=(triangleHalfLen*triangleHalfLen)/(bbox.width/2);
    this.markDistanceTextCenter=-this.innerR-(triangleHeight-bbox.width/2);
    Text.attr('y',this.markDistanceTextCenter).text('0.0');
    mark.insert('path','#markdistance')
      .attr('d', "M 0,"+(-this.innerR)+" l "+(-triangleHalfLen)+" "+(-triangleHeight)+" "+(2*triangleHalfLen)+" 0 z")
      .attr('class', 'mark');
    mark.insert('circle','#markdistance').attr('cx', '0').attr('cy', this.markDistanceTextCenter)
      .attr('r', bbox.width/2).attr('class', 'markBackground');

    Text=mark.append('text')
      .attr('id','marktext')
      .attr('x', '0').attr('y', -this.innerR).attr('class', 'marktext')
      .attr('text-anchor', 'middle')
      .attr('dy', '.35em') // IE does not support 'dominant-baseline' so use dy instead
      .text('i000i');
    this.svgMarkText=Text;
    bbox=Text.node().getBBox();
    this.markTextCenter=-this.innerR+bbox.width/2;
    Text.attr('y',this.markTextCenter).text('0');
    mark.insert('circle','#marktext').attr('cx', '0').attr('cy', this.markTextCenter)
      .attr('r', bbox.width/2).attr('class', 'markBackground');
    mark.attr("display", "none");
  },
  //************************************************************************************
  drawBackground: function (svgContainer, size) {
    var defs = svgContainer.append('defs');
    var gradient = defs.append('linearGradient')
      .attr('id', 'rosegradient')
      .attr('x1', '0')
      .attr('x2', '0')
      .attr('y1', '0')
      .attr('y2', '1')
    gradient.append('stop').attr('offset', '0%').attr('stop-color', '#0c5da5');
    gradient.append('stop').attr('offset', '100%').attr('stop-color', '#ff4f00');
    return svgContainer;
  },
  //************************************************************************************
  drawDepthIndicator: function (chart) {
    this.svgDepthText=chart.append('text')
      .attr('id', 'depth')
      .attr('x', '400').attr('y', '-340')
      .attr('text-anchor', 'end').attr('dominant-baseline', 'text-after-edge')
      .attr('class', 'positivegaugetext')
      .append('tspan')
      .attr('dy', '0px').attr('font-size', '60').text('99.9');
    chart.append("g").attr("transform","translate(300 -320)")
      .append("path").attr("id","depthSpark").attr("class","sparkline").attr("d","M 5,5 l 10,10 -10,0 z");
  },
  //************************************************************************************
  drawTrackLabel: function (chart) {
    var Text=chart.append('text')
      .attr('id', 'tracktruetext')
      .attr('x', 0).attr('y', 0)
      .attr('text-anchor', 'middle') //.attr('dominant-baseline', 'middle')
//      .attr('dy', '.3em') // IE does not support 'dominant-baseline' so use dy instead
      .attr('class', 'tracktext')
      .text('00000');
    this.svgTrackLabel=Text;
    SetElementPosition(Text,this.trackLabelPos,'#sailgaugeRect');
    var bbox=Text.node().getBBox();
    Text.text('');
    chart.insert('rect','#tracktruetext').attr('id', 'trackLabel')
      .attr('x', bbox.x).attr('y', bbox.y) //(this.trackLabelX)).attr('y', (this.trackLabelY))
      .attr('rx', (bbox.height*0.15)).attr('ry', (bbox.height*0.15))
      .attr('width', bbox.width).attr('height',bbox.height) //this.trackLabelRectW).attr('height', this.trackLabelRectH)
      .attr('class', 'tracktextbox');
  },
  //************************************************************************************
  drawSpeedLabel: function (chart) {
    var Text;
    var bbox;
    var Text=chart.append("text")
      .attr("id", "speed")
      .attr("x", 0).attr("y", 0)
      .attr("text-anchor", "middle") //.attr("dominant-baseline", "middle")
      .attr('dy', '.3em') // IE does not support 'dominant-baseline' so use dy instead
      .attr('class', 'speedtext')
      .text("00p00");
    this.svgSpeedLabel=Text;
    SetElementPosition(Text,this.speedLabelPos,'#sailgaugeRect');
    bbox=Text.node().getBBox();
    Text.text('');
    var r=chart.insert('rect','#speed').attr("id", "speedLabel")
      .attr('x', bbox.x).attr('y', bbox.y)
      .attr('rx', (bbox.height*0.15)).attr('ry', (bbox.height*0.15))
      .attr('width', bbox.width).attr('height',bbox.height)
      .attr('class', 'speedtextbox');
  },
  //************************************************************************************
  drawWindSectors: function (chart) {
    chart.append("g").attr("id","gportwindsector").attr("transform","rotate(0 0 0)")
      .append("path")
        .attr("id","portwindsector").attr("d","M0,0 v"+(-this.innerR)+" a"+(this.innerR)+","+(this.innerR)+" 1 0,1 0,0 z")
        .attr("fill","#ffdddd").attr("stroke","none");
    chart.append("g").attr("id","gstarboardwindsector").attr("transform","rotate(0 0 0)")
      .append("path")
        .attr("id","starboardwindsector").attr("d","M0,0 v"+(-this.innerR)+" a"+(this.innerR)+","+(this.innerR)+" 1 0,1 0,0 z")
        .attr("fill","#ddffdd").attr("stroke","none");
  },
  //************************************************************************************
  drawSvg: function (svgContainer, size) {
    this.drawBackground(svgContainer, size);
    var chart=svgContainer.append('g').attr('id', this.ID).attr('class', this.className); // add plain group box with this ID
    chart=chart.append('g').attr("transform", "rotate(0,"+this.center.x+","+this.center.y+") translate("+this.center.x+","+this.center.y+") scale("+this.scale+")");
    // append dummy rect for outer dimensions
    chart.append('rect').attr('id', this.SubID('Rect')).attr('x', -this.r).attr('y', -this.r).attr('width', this.r*2).attr('height',this.r*2).attr("fill","none").attr("stroke","none");
    this.svgRose = this.drawCompassRose(chart, size);
    this.drawTicks();
    this.drawLaylines(chart);
    this.drawBoat(chart);
    this.drawCourceIndicator(chart);
    this.drawMark(this.svgRose);
    this.drawTrueWindMarker(chart);
    this.drawApparentWindMarker(chart);
    this.drawSpeedLabel(chart);
    this.drawTrackLabel(chart);
    this.drawWindSectors(this.svgRose);
    this.drawDepthIndicator(chart);
  },
  //************************************************************************************
  drawCourceIndicator: function (svg) {
    this.svgCourceIndicator=svg.append('g');
    this.svgCourceIndicator.append("path")
      .attr("d","M 0 "+(-this.innerR-this.mainTickSize)+" l 0,"+(this.mainTickSize*10))
      .attr("class","courseIndicator");
  },
  //************************************************************************************
  drawTicks: function () {
    var tickmarks = d3.select("#tickmarks");
    var tickmarksTop = d3.select("#tickmarksTop");
    var innerR=this.innerR;
    
    for (var i = 0; i < 360; i += 10) {
      tickmarks.append("path").attr("id", "id10-" + i)
        .attr("d", "m 0,"+(innerR)+" L 0,"+(innerR+this.mainTickSize))
        .attr("class", "MainTick")
        .attr("transform", "rotate(" + i + " 0 0)");
    }
    for (var i = 5; i < 360; i += 10) {
      tickmarks.append("path").attr("id", "id5-" + i)
        .attr("d", "m 0,"+(innerR)+" L 0,"+(innerR+this.minorTickSize))
        .attr("class", "MinorTick")
        .attr("transform", "rotate(" + i + " 0 0)")
    }
    
    var dirs = ['N', 'E', 'S', 'W' ];
    // Add dummy text to get bounding box
    var cText=tickmarksTop.append("text").attr("id", "bearing")
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("class", "CardinalText")
        .text('INEI');
    var cbox=cText.node().getBBox();
    cText.remove();

    for (var i = 0; i < 360; i += 90) {
      tickmarksTop.append("rect").attr("id", "box" + i)
        .attr("x", cbox.x)
        .attr("y", cbox.y)
        .attr("width", cbox.width)
        .attr("height", cbox.height)
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DirBox")
        .text(i);
      tickmarksTop.append("text").attr("id", "bearing" + i)
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "CardinalText")
        .text(dirs[0]);
      dirs = dirs.splice(1);
    }
    var dirs = ['NE', 'SE', 'SW', 'NW'];
    // Add dummy text to get bounding box
    var icText=tickmarksTop.append("text").attr("id", "bearing")
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("class", "IntercardinalText")
        .text('IWWI');
    var icbox=icText.node().getBBox();
    icText.remove();

    for (var i = 45; i < 360; i += 90) {
      tickmarks.append("rect").attr("id", "box" + i)
        .attr("x", icbox.x)
        .attr("y", icbox.y)
        .attr("width", icbox.width)
        .attr("height", icbox.height)
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DirBox")
        .text(i);
      tickmarks.append("text").attr("id", "bearing" + i)
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "IntercardinalText")
        .text(dirs[0]);
      dirs = dirs.splice(1);
    }
  },
  //************************************************************************************
  drawSparkline: function (selector, data, width, height) {
    if (data.length > 0) {
      var now = Date.now();
      var x = d3.scale.linear().domain([
        d3.min(data, function (d) { return d.timestamp - now;}),
        d3.max(data, function (d) {return d.timestamp - now;})])
        .range([0, width]);
      var y = d3.scale.linear().domain([
        d3.min(data, function (d) {return d.value}),
        d3.max(data, function (d) {return d.value})])
        .range([0, height]);
      var line = d3.svg.line()
        .x(function (d) {return x(d.timestamp - now);})
        .y(function (d) {return y(d.value);});
      d3.select(selector).attr("d", line(data));
    }
  },
  //************************************************************************************
  drawBoat: function (chart) {
    var boatScale=this.innerR/300*2.5;
    this.svgBoat = chart.append('g').attr('id', this.SubID('Boat')).attr("display","none"); 
    var boat = this.svgBoat.append('g').attr('transform', 'rotate(90 0 0) translate(0 0) scale('+boatScale+')')
    boat.append('path')
      .attr('style', 'stroke:#00ff00;stroke-width:2px;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1')
      .attr('d', 'M -100,0 c 150,-100 200,-35 200,-35 0,35 0,35 0,35')
      .attr('class', 'boat sp');
    boat.append('path')
      .attr('style', 'stroke:#dd0000;stroke-width:2px;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1')
      .attr('d', 'M 100,0 c 0,35 0,35 0,35  0,0 -70,65 -200,-35')
      .attr('class', 'boat');
  },
  //************************************************************************************
  depthFontSize: function (depth) {
    var minFontSize = 60;
    var maxFontSize = 300;
    var shallowThreshold = 6;
    var minThreshold = 3;
    if (depth > shallowThreshold) {
      return minFontSize
    }
    if (depth < minThreshold) {
      return maxFontSize;
    }
    return minFontSize + (shallowThreshold - depth) / (shallowThreshold - minThreshold) * (maxFontSize - minFontSize);
  },
  //************************************************************************************
  arcForAngle: function (angle) {
    return 'M0,0 v-300 a300,300 1 0,1 ' +
      Math.sin(angle * Math.PI / 180) * 300 + ',' +
      (1 - Math.cos(angle * Math.PI / 180)) * 300 + ' z';
  },
  //************************************************************************************
  rotateAnimated: function (selector, angleFrom, angleTo, x, y, millis) {
    this.rotatesvgAnimated(d3.select(selector), angleFrom, angleTo, x, y, millis);
  },
  //************************************************************************************
  rotatesvgAnimated: function (d3g, angleFrom, angleTo, x, y, millis) {
    var rotateCenterStr=" " + x + " " + y;
    if ( this.useAnimatedRotation ) {
    var translateAndScaleStr="";//this.translateAndScaleStr;
    angleTo=angleTo % 360.0;
    var tween = function (d, i, a) {
       var aDist=angleTo-angleFrom;
       if ( Math.abs(aDist)>180 ) aDist=(aDist>0?aDist-360:aDist+360);
       var i = function(t) { 
         return angleFrom+aDist*t; 
       }
            return function(t) {
                return 'rotate(' + (i(t)) + "," + rotateCenterStr + ") "+translateAndScaleStr;
            };
    }
    d3g.transition()
      .duration(this.visible ? millis : 0)
      .attrTween('transform', tween);
    } else {
      d3g.attr("transform", "rotate(" + (angleTo) + " " + rotateCenterStr + ")");
    }
  }
}

//************************************************************************************
function DegToRad(degrees) {
  return degrees*Math.PI/180.0;
}

//************************************************************************************
function RadToDeg(radians) {
  return radians*180.0/Math.PI;
}

//************************************************************************************
function KnToms(knots) {
  return knots*1852.0/3600.0;
}

//************************************************************************************
// appAngle in radians
// appSpeed in m/s
// SpeedOverGround in m/s
// CourseOverGround in radians
// Drift in radians
function CalcTrueWind(appSpeed,appAngle,SpeedOverGround,Drift) {
  var angle=appAngle+Drift;
  var Vty = appSpeed*Math.sin(angle);
  var Vtx = appSpeed*Math.cos(angle) - SpeedOverGround;
  
  return { angle: Math.atan2(Vty,Vtx),
           speed: Math.sqrt(Math.pow(Vtx, 2)+ Math.pow(Vty, 2))
         };
}
