/*
  Copyright Timo Lappalainen, 2016
  
  Compass is simple svg based component for showing heading by using compass rose.
  Component is configurable by .css with class name.
  Methods:
    Compass(options)  - constructor
        options        - object of optional parameters. Define it as {ID:'compass2',className:'grayCompass'}
                         Optional parameters are;
          ID             - svg group id. So this can be searched with d3.select(#<ID>);
          className      - defines class name for .css . Default 'compass'
          margin         - margin around rose box. Default 10
          mainTickSize   - Size of 10 degree tics. Default 15.
          minorTickSize  - Size of 5 degree tics. Default 8.
          HeadingFilter  - Minimum value for rotating rose. This prevents small jitter. Default 0.5.
          
    init(svgContainer,x,y,boxWidth,scale) - initialize and draw componenent
        svgContainer   - svg container
        options        - object of optional parameters. Define it as {x:200, y:200}
                         Optional parameters are;
          center:{x,y} - Object of center of compass rose. Define as center:{x:500,y:500}. 
                         These are container coordinates. Default is center of container.
          size           - size of compass in container units. Default is min size of container.
          
    updateHeading(value) - set heading and rotates compass.
    
  Properties:
    All contructor options and -
    svgContainer       - svgContainer where node stays
    scale              - scale of the componenent
    center             - Object of center of the rose. center has properties x,y.
  
  Styles, using default class name "compass" (see className property)
    .compass {
    }
    .compass .rose {stroke:#A63400;fill:#679fd2}
    .compass .rose .text{stroke:#A63400;fill:#A63400}
    .compass .rose .MainTick{stroke:A63400; stroke-width:2; fill:A63400}
    .compass .rose .MinorTick{stroke:A63400; stroke-width:1; fill:A63400}
    .compass .rose .OuterCircle{stroke:red; stroke-width:1; fill:white}
    .compass .rose .InnerCircle{stroke:red; stroke-width:1; fill:gray}
    .compass .rose .CardinalText{fill:#A63400; stroke-width:1 ;font-size:200%}
    .compass .rose .IntercardinalText{stroke:#A63400; stroke-width:1 ;fill:#A63400;font-size:160%}
    .compass .rose .DiamondCardinalCW{stroke:red; fill:red}
    .compass .rose .DiamondCardinalCCW{stroke:black; fill:black}
    .compass .rose .DiamondIntercardinalCW{stroke:red; fill:red}
    .compass .rose .DiamondIntercardinalCCW{stroke:black; fill:black}
    .compass .rose .DirBox {fill:#679fd2}
    .compass .SteerMark {fill:red}

  Sample usage:
  ...
	<body>
    ...
    <div id="DisplayContainer" class="svg"></div>
    ...
    <script>
      var svgContainer=AddSvgContainer('#DisplayContainer',1000);
      var compass = new Compass();
      compass.init(svgContainer,{center:{x:500,y:500}});
  
*/

//************************************************************************************
function Compass(options) {
  var def={ID:null
           ,className:'compass'
           ,margin:10
           ,mainTickSize:15
           ,minorTickSize:8
           ,HeadingFilter:0.5
          }; 
  if (options==null) options=def;
  CombineDefaultsToObject(options,def);
  CombineDefaultsToObject(this,options);

  if ( this.ID==null ) this.ID='Compass';
  this.Heading = 0;
  this.RoseHeading = 0; // this is actual rose heading. It may differ from Heading with value of HeadingFilter.
  this.margin=DefaultNum(options.margin,def.margin);
  this.mainTickSize=DefaultNum(options.mainTickSize,def.mainTickSize);
  this.minorTickSize=DefaultNum(options.minorTickSize,def.minorTickSize);
  this.HeadingFilter=DefaultNum(options.HeadingFilter,def.HeadingFilter);
}

//************************************************************************************
Compass.prototype = {
    init: function (svgContainer,options) {
    var def={ID:null
            ,center:{x:null,y:null} 
            ,size:null
            }; 
    if (options==null) options=def;
    options=CombineDefaultsToObject(options,def);

    this.r=200;
    this.svgContainer=svgContainer;
    var svgSizeInfo=GetSvgSizeInfo(svgContainer);
    options.center.x=DefaultNum(options.center.x,svgSizeInfo.width/2);
    options.center.y=DefaultNum(options.center.y,svgSizeInfo.height/2);
    var defSize=Math.min(svgSizeInfo.width,svgSizeInfo.height);
    options.size=DefaultNum(options.size,defSize);
    this.scale=options.size/2/this.r;
    this.center=options.center;
    this.waitInitDraw=true;
    this.drawCompass(svgContainer);
    this.Visible=false;
    this.Warning="N/A";
    this.startForceVisible();
  },
  //************************************************************************************
  updateHeading: function (value) {
    if ( !isNaN(value) ) {
      this.Heading=value;
      if ( Math.abs(this.RoseHeading-value)>this.HeadingFilter ) {
        this.rotateAnimated(value, 100);
      }
      this.headingVisible(true);
      this.showWarning();
    } else {
      this.showWarning(value);
    }  
  },
  //************************************************************************************
  showWarning: function (val) {
    if ( val!=this.Warning ) {
      if ( val==null ) {
        this.WarnText.attr("display","none");
      } else {
        this.WarnText.text(val).attr("display", "inline");
      }
      this.Warning=val;
    }
  },
  //************************************************************************************
  headingVisible: function (val) {
    if ( val!=this.Visible ) {
      var visible="none";
      if ( val ) visible="inline";
      this.diamond.attr("display", visible);
      this.tickmarks.attr("display", visible);
      this.tickmarksTop.attr("display", visible);
      this.Visible=val;
    }
  },
  //************************************************************************************
  startForceVisible: function () {
    var that = this;
    setTimeout(function showHeading() {
      that.headingVisible(true);
    }, 1000);
  },
  //************************************************************************************
  drawCompass: function (svg) {
    this.rose = this.drawCompassRose(svg);
    this.drawTicks();
//    this.drawMark(rose);
  },
  //************************************************************************************
  drawCompassRose: function (svg) {
    this.rotateCenterStr=this.center.x+","+this.center.y;
    this.translateAndScaleStr="translate("+this.center.x+","+this.center.y+") scale("+this.scale+")";
    var compass=svg.append('g').attr('id', this.ID).attr('class', this.className);
    var rose = compass.append('g').attr('id', this.ID+"_rose").attr("transform", "rotate(0,"+this.center.x+","+this.center.y+") translate("+this.center.x+","+this.center.y+") scale("+this.scale+")").attr('class', 'rose');
    var mark = compass.append('g').attr('id', this.ID+"_mark").attr("transform", "rotate(0,"+this.center.x+","+this.center.y+") translate("+this.center.x+","+this.center.y+") scale("+this.scale+")");
    mark.append('path')
      .attr('d', 'm 0,'+(-(this.r-this.margin-this.mainTickSize+this.minorTickSize))+' l '+(this.mainTickSize/2)+','+(-this.mainTickSize)+' h '+(-this.mainTickSize)+' z')
      .attr('class', 'SteerMark');
    rose.append('g').attr('id', 'edge').attr('class', 'p1');
    rose.append('circle').attr('cx', 0).attr('cy', 0).attr('r', this.r - this.margin)
        .attr("class", "OuterCircle");
    rose.append('circle').attr('cx', 0).attr('cy', 0).attr('r', this.r - this.margin-this.mainTickSize)
        .attr("class", "InnerCircle");
    rose.append('g').attr('id', 'tickmarks');
    rose.append('g').attr('id', 'tickmarksTop');
    rose.append('g').attr('id', 'diamond');
    this.WarnText=mark.append('text')
          .attr("x", 0)
          .attr("y", 0)
          .attr("text-anchor", "middle")
          .attr('dy', '0.3em')
          .attr("class", "WarnText")
          .attr("display","none")
          .text("N/A");
    return rose;
  },
  //************************************************************************************
  drawTicks: function () {
    var RS=5;
    var DiamondCardinalCW=[ {"x": 0*RS, "y": 0*RS}, {"x": 0*RS, "y":-28*RS}, {"x": 4*RS, "y": -8*RS} ];
    var DiamondCardinalCCW=[ {"x": 0*RS, "y": 0*RS}, {"x": 0*RS, "y":-28*RS}, {"x": -4*RS, "y": -8*RS} ];
    var DiamondIntercardinalCW=[ {"x": 0*RS, "y": 0*RS}, {"x":-14*RS, "y":-14*RS}, {"x": -4*RS, "y": -8*RS} ];
    var DiamondIntercardinalCCW=[ {"x": 0*RS, "y": 0*RS}, {"x":-14*RS, "y":-14*RS}, {"x": -8*RS, "y": -4*RS} ];

    this.tickmarks = d3.select("#tickmarks");
    this.tickmarksTop = d3.select("#tickmarksTop");
    this.diamond = d3.select("#diamond");
    var innerR=this.r - this.margin-this.mainTickSize;
    
    var RoseFunction=d3.svg.line()
                       .x(function(d) { return d.x; }) 
                       .y(function(d) { return d.y; })
                       .interpolate("linear");

    for (var i = 0; i < 360; i += 10) {
      this.tickmarks.append("path").attr("id", "id10-" + i)
        .attr("d", "m 0,"+(innerR)+" L 0,"+(innerR+this.mainTickSize))
        .attr("class", "MainTick")
        .attr("transform", "rotate(" + i + " 0 0)")
    }
    for (var i = 5; i < 360; i += 10) {
      this.tickmarks.append("path").attr("id", "id5-" + i)
        .attr("d", "m 0,"+(innerR)+" L 0,"+(innerR+this.minorTickSize))
        .attr("class", "MinorTick")
        .attr("transform", "rotate(" + i + " 0 0)")
    }
    var dirs = ['N', 'E', 'S', 'W' ];
    // Add dummy text to get bounding box
    var cText=this.tickmarksTop.append("text").attr("id", "bearing")
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("class", "CardinalText")
        .text('INEI');
    var cbox=cText.node().getBBox();
    cText.remove();

    for (var i = 0; i < 360; i += 90) {
      this.diamond.append("path").attr("id", "diamond" + i)
        .attr("d", RoseFunction(DiamondCardinalCW))
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DiamondCardinalCW");
      this.diamond.append("path").attr("id", "diamond" + i)
        .attr("d", RoseFunction(DiamondCardinalCCW))
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DiamondCardinalCCW");
      this.diamond.append("path").attr("id", "diamondS" + i)
        .attr("d", RoseFunction(DiamondIntercardinalCW))
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DiamondIntercardinalCW");
      this.diamond.append("path").attr("id", "diamond" + i)
        .attr("d", RoseFunction(DiamondIntercardinalCCW))
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DiamondIntercardinalCCW");
      this.tickmarksTop.append("rect").attr("id", "box" + i)
        .attr("x", cbox.x)
        .attr("y", cbox.y)
        .attr("width", cbox.width)
        .attr("height", cbox.height)
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DirBox")
        .text(i);
      this.tickmarksTop.append("text").attr("id", "bearing" + i)
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "CardinalText")
        .text(dirs[0]);
      dirs = dirs.splice(1);
    }
    var dirs = ['NE', 'SE', 'SW', 'NW'];
    // Add dummy text to get bounding box
    var icText=this.tickmarksTop.append("text").attr("id", "bearing")
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("class", "IntercardinalText")
        .text('IWWI');
    var icbox=icText.node().getBBox();
    icText.remove();

    for (var i = 45; i < 360; i += 90) {
      this.tickmarks.append("rect").attr("id", "box" + i)
        .attr("x", icbox.x)
        .attr("y", icbox.y)
        .attr("width", icbox.width)
        .attr("height", icbox.height)
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "DirBox")
        .text(i);
      this.tickmarks.append("text").attr("id", "bearing" + i)
        .attr("x", 0)
        .attr("y", (-innerR))
        .attr("text-anchor", "middle")
        .attr('dy', '1em')
        .attr("transform", "rotate(" + i + " 0 0)")
        .attr("class", "IntercardinalText")
        .text(dirs[0]);
      dirs = dirs.splice(1);
    }
    this.headingVisible(false);
  },
  //************************************************************************************
  rotateAnimated: function (angleTo, millis) {
    var angleFrom=this.Heading;
    var rotateCenterStr=this.rotateCenterStr;
    var translateAndScaleStr=this.translateAndScaleStr;
    angleTo=angleTo % 360.0;
    var tween = function (d, i, a) {
       var aDist=angleTo-angleFrom;
       if ( Math.abs(aDist)>180 ) aDist=(aDist>0?aDist-360:aDist+360);
       var i = function(t) { 
         return angleFrom+aDist*t; 
       }
            return function(t) {
                return 'rotate(' + (-i(t)) + "," + rotateCenterStr + ") "+translateAndScaleStr;
            };
    }
    this.rose.transition()
      .duration(this.visible ? millis : 0)
      .attrTween('transform', tween);
    this.RoseHeading=angleTo;
  },
}
