/*
  Copyright Timo Lappalainen, 2016
  
  ValueBox is simple svg based component for showing value and it's name above it.
  Component is configurable by .css with class name.
  Methods:
    ValueBox(options)  - constructor
        options        - object of optional parameters. Define it as {sampleValue:'15,6 kn',valueFormat:'%.1f kn', ID='MyBox'}
                         Optional parameters are;
          sampleValue    - sample value for calculating box size
          valueName      - name of value printed above value box. If empty, no name will be printed. Default ""
          ID             - svg group id. So this can be searched with d3.select(#<ID>);
          className      - defines class name for .css . Default 'valueBox'
          frameSize      - frame around value name and value. Default 10
          nameMargin     - margin for value name on start/end align. Default 10.
          valueMargin    - margin for value name on start/end align. Default 10.
          valueFormat    - sprintf style format for value given as number with updateValue
          
    init(svgContainer,options) - initialize and draw componenent
        svgContainer   - svg container
        options        - object of optional parameters. Define it as {x:200, y:200, boxWidth:400}
                         Optional parameters are;
          x,y            - x,y for top,left corner of componenet on svg container.
                           these are container coordinates
          boxWidth       - width of box in container units. Note that this also scales
                           component. The height will be defined by font sizes on styles
          scale          - component scaling. Default scale is boxWidth/400;
          
    updateStr(value)   - set string value on value box.
    
    updateValue(value) - set value by using sprintf(valueFormat,value)
    
  Properties:
    All contructor and init options and -
    svgContainer       - svgContainer where node stays
    scale              - scale of the componenent
    left               - left coordinate of the component in svg container units
    top                - top coordinate of the component in svg container units
    width              - width of the component in svg container units
    height             - height of the component in svg container units
    
  
  Styles, using default class name "valueBox" (see className property)
    .valueBox {
    }
    .valueBox .FrameBox{stroke:#A63400; stroke-width:3;fill:white}
    .valueBox .ValueNameBox{stroke:#A63400; stroke-width:2;fill:black}
    .valueBox .ValueNameText{stroke:white; fill:white; text-anchor:start; stroke-width:1 ;font-size:300%}
    .valueBox .ValueBox{stroke:#A63400; stroke-width:2;fill:black}
    .valueBox .ValueText{stroke:white; text-anchor:middle; fill:white; stroke-width:1 ;font-size:600%}

  Aligning boxes:
    There are several ways to define box position.
    - As default top, left and with container width.
        speedBox.init(svgContainer);
    - Define hard x,y,size
        speedBox.init(svgContainer,{x:10,y:200,boxWidth:300});
    - Referenced to container
        speedBox.init(svgContainer,{x:GetSvgSizeInfo(svgContainer).center.x
                                   ,y:GetSvgSizeInfo(svgContainer).center.y
                                   ,boxWidth:300});
    - Aligning to given value
        speedBox.init(svgContainer,{x:"center:"+GetSvgSizeInfo(svgContainer).center.x
                                   ,y:"below:"+GetSvgSizeInfo(svgContainer).center.y, dy:15
                                   ,boxWidth:300});
    - Aligning to other node
        speedBox.init(svgContainer,{x:"center:#"+compass.ID
                                   ,y:"below:#"+compass.ID, dy:15
                                   ,boxWidth:300});
        depthBox.init(svgContainer,{x:"center:#SpeedBox"
                                   ,y:"below:#SpeedBox", dy:15
                                   ,boxWidth:speedBox.boxWidth
                                   ,scale:speedBox.scale
                                   });
  
  Sample usage:
  ...
	<body>
    ...
    <div id="DisplayContainer" class="svg"></div>
    ...
    <script>
      var svgContainer=AddSvgContainer('#DisplayContainer',1000);
      var speedBox = new ValueBox('15,6 kn','Speed','SpeedBox');
      var depthBox = new ValueBox('150,2 m','Depth','DepthBox');
      speedBox.valueFormat="%.1f kn";
      depthBox.valueFormat="%.1f m";
      speedBox.init(svgContainer,200,15,600);
      depthBox.init(svgContainer,{x:"left:#batteryVoltage"
                                 ,y:"below:#batteryVoltage", dy:15
                                 ,boxWidth:batteryVoltage.boxWidth
                                 ,scale:batteryVoltage.scale
                                 });
    </script>
  </body>
  
*/

var ValueBoxIDIndex=0;

//************************************************************************************
function ValueBox(options) {
  var def={sampleValue:null
           ,valueName:null
           ,ID:null
           ,className:'valueBox'
           ,frameSize:10
           ,nameMargin:10
           ,valueMargin:10
           ,valueFormat:'%.1f'
           ,frameRounding:'15'
           ,innerFrameRounding:null
          }; 
  if (options==null) options=def;
  CombineDefaultsToObject(options,def);
  CombineDefaultsToObject(this,options);

  if ( this.ID==null ) {
    ID='ValueBox'+IDIndex;
    IDIndex++;
  }
  this.ID=CreateSafeID(this.ID);
  
  if (this.valueName==null) this.valueName='';
  if ( this.sampleValue==null ) this.sampleValue='N/A';
  
  this.frameSize=DefaultNum(options.frameSize,def.frameSize);
  this.nameMargin=DefaultNum(options.nameMargin,def.nameMargin);
  this.valueMargin=DefaultNum(options.valueMargin,def.valueMargin);
  if ( this.innerFrameRounding==null ) {
    this.innerFrameRounding=this.frameRounding;
    // Is there better way to calculate with units?
    if ( isNaN(this.innerFrameRounding) ) {
      var s = this.innerFrameRounding.replace(/[^\d.-]/g, ''); // remove non chars
      var val=parseFloat(s);
      this.innerFrameRounding=(val*0.76).toString()+this.innerFrameRounding.replace(s,'');
    } else this.innerFrameRounding*=0.76;
  }
}

//************************************************************************************
ValueBox.prototype = {
  init: function (svgContainer,options) {
    var def={x:null
            ,y:null
            ,dx:0
            ,dy:0
            ,boxWidth:null
            ,scale:null
            };
    if (options==null) options=def;
    CombineDefaultsToObject(options,def);
    CombineDefaultsToObject(this,options);

    this.svgContainer=svgContainer;
    var svgSizeInfo=GetSvgSizeInfo(svgContainer);
    var defSize=Math.min(svgSizeInfo.width,svgSizeInfo.height);
    this.boxWidth=DefaultNum(this.boxWidth,defSize);
    this.scale=DefaultNum(this.scale,this.boxWidth/400);
    this.dx=DefaultNum(this.dx,0);
    this.dy=DefaultNum(this.dy,0);
    
    this.size=this.boxWidth/this.scale;
    this.left=DefaultNum(this.x+this.dx,(svgSizeInfo.width-this.boxWidth)/2);
    this.top=DefaultNum(this.y+this.dy,svgSizeInfo.height/2);
    this.drawBox(svgContainer);
    this.moveBox(this.x,this.y);
    
    this.Value=0;
    this.strValue="";
  },

  //************************************************************************************
  updateStr: function (value) {
    if ( value==null ) value='null';
    this.Select("ValueText").text(value);
    this.strValue=value;
  },

  //************************************************************************************
  updateValue: function (value) {
    if ( !isNaN(value) ) {
      this.updateStr(sprintf(this.valueFormat,value));
      this.Value=value;
    } else {
      this.updateStr(value);
    }
  },

  //************************************************************************************
  SubID: function (ID) {
    return this.ID+ID;
  },
  //************************************************************************************
  Select: function (ID) {
    return d3.select("#"+this.SubID(ID));
  },
  //************************************************************************************
  moveBox: function (x,y) {
    this.x=x; this.y=y;
    if ( isNaN(this.x) ) { // try to resolve
      x=CalculateElementXPos(this.Select(""),this.x,"");
    }
    if ( isNaN(this.y) ) { // try to resolve
      y=CalculateElementYPos(this.Select(""),this.y,"");
    }
    this.left=x+this.dx;
    this.top=y+this.dy;
    this.Select("Box").attr("transform", "translate("+this.left+","+this.top+") scale("+this.scale+")");
  },
  //************************************************************************************
  drawBox: function (svg) {
    // First add untransformed group for reading bound box in main svg units.
    var boxMain = svg.append('g').attr('id', this.ID).attr('class', this.className);
    var box = boxMain.append('g').attr('id', this.SubID("Box")).attr("transform", "translate("+this.left+","+this.top+") scale("+this.scale+")");

    var valueNamebox={height:0};
    var valueNameAlign="";
    var valueBoxTop=this.frameSize;
    var valueName;
    var valueText;
    
    if ( this.valueName != '' ) {
      // Create value name node
      var valueName=box.append("text").attr("id", this.SubID("ValueNameText"))
          .attr("x", this.frameSize+this.nameMargin)
          .attr("y", this.frameSize)
          .attr('dy', '0.9em')
          .attr("class", "ValueNameText")
          .text(this.valueName);
      valueNamebox=valueName.node().getBBox();
      valueBoxTop=this.frameSize*2+valueNamebox.height;
      valueNameAlign=valueName.style("text-anchor");
    }
    // Create value node
    valueText=box.append("text").attr("id", this.SubID("ValueText"))
        .attr("x", 0)
        .attr("y", valueBoxTop)
        .attr('dy', '0.9em')
        .attr("class", "ValueText")
        .text(this.sampleValue);
    var valueBox=valueText.node().getBBox();
    var valueAlign=valueText.style("text-anchor");
    
    // Insert frame node
    this.height=(valueBoxTop+valueBox.height+this.frameSize);
    box.insert("rect","#"+this.SubID("ValueNameText")).attr("id", this.SubID("FrameBox"))
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", this.size)
        .attr("height", this.height)
        .attr('rx', this.frameRounding).attr('ry', this.frameRounding)
        .attr("class", "FrameBox");
    
    if ( this.valueName != '' ) {
      // insert value name background rect node
      box.insert("rect","#"+this.SubID("ValueNameText")).attr("id", "ValueNameBox")
          .attr("x", this.frameSize)
          .attr("y", this.frameSize)
          .attr("width", this.size-this.frameSize*2)
          .attr("height", valueNamebox.height)
          .attr('rx', this.innerFrameRounding).attr('ry', this.innerFrameRounding)
          .attr("class", "ValueNameBox");
      var ValueNameX=this.size/2; // default middle align for text
      switch(valueNameAlign) {
          case 'start':
            ValueNameX=this.frameSize+this.nameMargin;
            break;
          case 'end':
            ValueNameX=this.size-(this.frameSize+this.nameMargin);
            break;
      }
      valueName.attr("x", ValueNameX);
    }
    
    // insert value background rect node
    box.insert("rect","#"+this.SubID("ValueNameText")).attr("id", this.SubID("ValueBox"))
        .attr("x", this.frameSize)
        .attr("y", valueBoxTop)
        .attr("width", this.size-this.frameSize*2)
        .attr("height", valueBox.height)
        .attr('rx', this.innerFrameRounding).attr('ry', this.innerFrameRounding)
        .attr("class", "ValueBox");
    var ValueX=this.size/2; // default middle align for text
    switch(valueAlign) {
        case 'start':
          ValueX=this.frameSize+this.valueMargin;
          break;
        case 'end':
          ValueX=this.size-(this.frameSize+this.valueMargin);
          break;
    }
    valueText.attr("x", ValueX);
  }
}
